import os, sys
from distutils import sysconfig
from distutils.ccompiler import new_compiler, show_compilers
from distutils.core import Command
from distutils.dep_util import newer

SCRIPT = """\
#!/usr/bin/env python
import sys
sys.path.insert(0, %(install_dir)r)
%(import)s
%(ident)s().run()
"""

WIN_EXEC = """\
#include <windows.h>
#include <string.h>
#include <process.h>
#include <stdio.h>

int main(int argc, char **argv) {
  char modulename[MAX_PATH+1];
  char *scriptname = "%(name)s";
  char **pyargv;
  char *p;
  int i, j;
  
  GetModuleFileName(NULL, modulename, MAX_PATH);

  /* remove the executable name */
  i = strlen(modulename);
  p = strrchr(modulename, '\\\\');
  if (p)
    /* save the slash */
    *(++p) = '\\0';
  else
    modulename[0] = '\\0';
  strcat(modulename, scriptname);

  /* create the new argv */
  pyargv = (char**) calloc(argc + 2, sizeof(char *));
  pyargv[0] = "%(executable)s";
  pyargv[1] = modulename;
  for (i = 1, j = 2; i < argc; i++, j++) {
    pyargv[j] = argv[i];
    }

  i = spawnv(P_WAIT, "%(executable)s", pyargv);
  free(pyargv);
  return i;
}
"""
  
  
class BuildScripts(Command):

    command_name = 'build_scripts'

    description = "\"build\" scripts"

    user_options = [
        ('build-dir=', 'd', "directory to \"build\" (copy) to"),
        ('force', 'f', "forcibly build everything (ignore file timestamps"),
        ]

    if os.name == 'nt':
        user_options.extend([
        ('debug', 'g', "compile/link with debugging information"),
        ('build-temp=', 't', "directory for temporary files (build by-products)"),
        ('compiler=', 'c', "specify the compiler type"),
        ])

        help_options = [
            ('help-compiler', None,
             "list available compilers", show_compilers),
            ]

    boolean_options = ['force']

    def initialize_options(self):
        self.build_dir = None
        self.scripts = None
        self.force = None
        self.outfiles = None
        # For scripts on Windows (actually executables)
        self.build_temp = None
        self.compiler = None
        self.debug = None

    def finalize_options(self):
        undefined_temp = self.build_temp is None

        self.set_undefined_options('build',
                                   ('build_scripts', 'build_dir'),
                                   ('build_temp', 'build_temp'),
                                   ('compiler', 'compiler'),
                                   ('debug', 'debug'),
                                   ('force', 'force'))

        self.scripts = self.distribution.scripts
        self.outfiles = []

        if undefined_temp:
            self.build_temp = os.path.join(self.build_temp, 'scripts')
        return

    def run(self):
        """
        Create the proper script for the current platform.
        """
        if not self.scripts:
            return

        install_lib = self.get_finalized_command('install_lib')
        install_dir = install_lib.install_dir

        self.mkpath(self.build_dir)

        if os.name == 'nt':
            self.create_compiler()

        for script in self.scripts:

            # Always build the scripts as the install_dir might be changed
            self.announce("building '%s' script" % script.name)

            outfile = os.path.join(self.build_dir, script.name)
            self.outfiles.append(outfile)

            if not self.dry_run:
                args = {'ident' : script.identifier,
                        'install_dir' : install_dir}
                
                if script.module:
                    args['import'] = 'from %s import %s' % (script.module, 
                                                            script.identifier)
                else:
                    args['import'] = 'import %s' % script.identifier

                f = open(outfile, 'w')
                f.write(SCRIPT % args)
                f.close()

            if os.name == 'nt':
                py_base, py_ext = os.path.splitext(sys.executable)
                if py_base.endswith('_d'):
                    py_base = py_base[:-2]
                    
                if self.debug:
                    # Windows debug build
                    progname = script.name + '_d'
                    py_exec = py_base + '_d' + py_ext
                else:
                    progname = script.name
                    py_exec = py_base + py_ext

                outfile = os.path.join(self.build_dir, progname + '.exe')
                self.outfiles.append(outfile)

                if not self.dry_run:
                    args = {'executable' : py_exec.replace('\\', '\\\\'),
                            'name' : script.name,
                            }

                    source = os.path.join(self.build_temp, script.name + '.c')
                    f = open(source, 'w')
                    f.write(WIN_EXEC % args)
                    f.close()

                    # Now compile and link it
                    objects = self.compiler.compile([source],
                                                    output_dir=self.build_temp,
                                                    debug=self.debug,
                                                    )
                    self.compiler.link_executable(objects, progname,
                                                  output_dir=self.build_dir,
                                                  debug=self.debug,
                                                  )
        return

    # -- windows scripts --------------------------------------------
    
    def create_compiler(self):
        # Ensure by-product directory exists
        self.mkpath(self.build_temp)
        
        # Setup the CCompiler object that we'll use to do all the
        # compiling and linking
        self.compiler = new_compiler(compiler=self.compiler,
                                     verbose=self.verbose,
                                     dry_run=self.dry_run,
                                     force=self.force)
        sysconfig.customize_compiler(self.compiler)

        # Make sure Python's include directories (for Python.h, pyconfig.h,
        # etc.) are in the include search path.
        py_include = sysconfig.get_python_inc()
        include_dirs = [py_include]

        plat_py_include = sysconfig.get_python_inc(plat_specific=1)
        if plat_py_include != py_include:
            include_dirs.append(plat_py_include)
        self.compiler.set_include_dirs(include_dirs)
    
        # Python's library directory must be in the library search path
        library_dirs = [os.path.join(sys.exec_prefix, 'libs')]
        self.compiler.set_library_dirs(library_dirs)
        return
