########################################################################
#
# File Name:            DocumentDefinition.py
#
# Documentation:        http://docs.4suite.org/4SS/DocumentDefinition.py.html
#
"""
4Suite server document definitions: contain system and user metadata for managed documents
WWW: http://4suite.org         e-mail: support@4suite.org

Copyright (c) 2001 Fourthought, Inc. USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import cPickle, cStringIO, sys, os

from Ft.Server.Server import FtServerServerException,Error

import XmlDocumentImp

from Ft.Server import FTSERVER_NAMESPACE

from Ft.Xml import XPath
from Ft.Xml.Domlette import Print

from Ft.Xml.XPath import Conversions, Context
from types import ListType
from Ft.Rdf import Statement
from Ft.Xml import Domlette

from Ft.Server.Common import ResourceTypes, Schema, DocumentDefinitionSerialization, CreationParams, XmlLib

from Ft.Server.Server.Drivers import FtssDriver
from Ft.Xml.XLink import XLINK_NAMESPACE


class DocumentDefinitionImp(XmlDocumentImp.XmlDocumentImp):
    resourceType = ResourceTypes.ResourceType.DOCUMENT_DEFINITION

    def addBaseDocumentDefinition(self, baseDef):
        """
        Add a base document definition to this document definition.
        This will cause all of the documents associated with this
        document definition to be re-processed and could take a long time
        """
        self._verifyTx()
        if not baseDef.isResourceType(ResourceTypes.ResourceType.DOCUMENT_DEFINITION):
            raise FtServerServerException(Error.INVALID_PATH,
                                          path = baseDef.getPath().displayPath,
                                          type='Document Definition')

        xu = NEW_BASE_XUPDATE%(FTSERVER_NAMESPACE, XLINK_NAMESPACE, baseDef.getPath().absolutePath)
        self._driver.xupdateContent(self.getPath(), xu)
        self._update()
        return

    def setContent(self, src):
        XmlDocumentImp.XmlDocumentImp.setContent(self, src)
        self._update()
        return

    def _delete(self):
        self._verifyTx()
        for pdd in self.getDerivedDefinitions():
            pdd.removeBaseDocumentDefinition(self)

        #Set all of our meta documents to have a None doc def
        for path in self.getMetaDocumentUris(recurse = 0):
            #Just do a reset of the content
            doc = self.fetchResource(path)
            doc.setDocumentDefinition(None)

        XmlDocumentImp.XmlDocumentImp._delete(self)
        return

    def getBaseDefinitions(self):
        """
        Get the base definitions of this document definition
        """
        self._verifyTx()
        paths = [ x.object for x in self._driver.getSystemModel().complete(self._path.absolutePath, Schema.BASE_DOCDEF, None) ]
        res = []
        for p in paths:
            path = self._basePath.normalize(p,1)
            res.append(self._fetchResource(path))
        return res
        
    def getCreationParams(self):
        """
        Get the creation parameters associated with this document definition.  See Ft.Server.Core.CreationParamsImp.py
        """
        self._verifyTx()
        context = self.toXPathContext()
        cp = CreationParams._Deserialize(context)
        return cp

    def getDerivedDefinitions(self):
        """
        Get all of the parent definitions of this document definition
        """
        paths = [ x.subject for x in self._driver.getSystemModel().complete(None, Schema.BASE_DOCDEF, self._path.absolutePath) ]
        res = []
        for p in paths:
            path = self._basePath.normalize(p, 1)
            res.append(self._fetchResource(path))
        return res

    def getMetaDocumentUris(self, recurse=1):
        """
        Get a list of all documents that use this doc def
        """
        res = []
        if recurse:
            for bd in self.getBaseDefinitions():
                res.extend(bd.getMetaDocumentUris())
        res.extend([ x.subject for x in self._driver.getSystemModel().complete(None, Schema.DOCDEF, self.getPath().absolutePath) ])
        return res

    def removeBaseDocumentDefinition(self, baseDef, recurse=1):
        """
        Remove a base document definition.  This will cause all of the documents
        associated with this document definition to be regenerated.
        """
        if not baseDef.isResourceType(ResourceTypes.ResourceType.DOCUMENT_DEFINITION):
            raise FtServerServerException(Error.INVALID_PATH,
                                          path = baseDef.getPath().displayPath,
                                          type='Document Definition')

        self._verifyTx()
        xu = REMOVE_BASE_XUPDATE%(FTSERVER_NAMESPACE, XLINK_NAMESPACE,
                                  baseDef.getPath().absolutePath)
        self._driver.xupdateContent(self.getPath(), xu)
        self._update()
        return

    def setCreationParams(self,newCp):
        """
        Set the creation parameters of this document definition
        """
        self._verifyTx()

        doc = Domlette.implementation.createDocument(None,None,None)
        cp = CreationParams._Serialize(doc,newCp)

        st = cStringIO.StringIO()
        Print(cp, stream=st)
        xu = SET_CREATION_PARAMS_XUPDATE%(FTSERVER_NAMESPACE, XLINK_NAMESPACE,
                                          st.getvalue())
        self._driver.xupdateContent(self._path, xu)
        self._update()
        return

    def _update(self):
        """This is called when ever we need to regenerate all of our meta documents"""
        #This process could take a long time
        #First Do our meta documents,
        #Then, do all of our derived definitions
        for path in self.getMetaDocumentUris(recurse = 0):
            #Just do a reset of the content
            doc = self.fetchResource(path)
            doc.setContent(doc.getContent())
        for dd in self.getDerivedDefinitions():
            dd._update()
        return


def NewDocumentDefinitionXml(driver, path, acl, owner, imt, src, docDef, typ):
    a = driver.aclToXml(acl)
    t = FtssDriver.CurrentTime()
    md = """<ftss:MetaData xmlns:ftss="%s" path='%s' type='%s' creation-date='%s' document-definition='%s'>
  %s
<ftss:LastModifiedDate>%s</ftss:LastModifiedDate>
<ftss:Owner>%s</ftss:Owner>
<ftss:Imt>%s</ftss:Imt>
<ftss:Size>%d</ftss:Size>
</ftss:MetaData>
        """ % (FTSERVER_NAMESPACE, path, Schema.g_rdfResourceTypes[typ],
               t, docDef or Schema.NULL_DOCDEF, a, t, owner, imt, len(src),
               )
    return XmlLib.MakeString(md)


NEW_BASE_XUPDATE="""<?xml version="1.0"?>
<xupdate:modifications
  version="1.0"
  xmlns:xupdate="http://www.xmldb.org/xupdate"
  xmlns:ftss="%s"
  xmlns:xlink="%s"
>
  <xupdate:append select="/*/ftss:BaseNames" child="last()">
    <ftss:Base xlink:type="simple" xlink:href="%s" xlink:actuate="onLoad" xlink:show="embed"/>
  </xupdate:append>

</xupdate:modifications>
"""

REMOVE_BASE_XUPDATE="""<?xml version="1.0"?>
<xupdate:modifications
  version="1.0"
  xmlns:xupdate="http://www.xmldb.org/xupdate"
  xmlns:ftss="%s"
  xmlns:xlink="%s"
>
<xupdate:remove select="/*/ftss:BaseNames/ftss:Base[@xlink:href='%s']"/>

</xupdate:modifications>
"""


SET_CREATION_PARAMS_XUPDATE="""<?xml version="1.0"?>
<xupdate:modifications
  version="1.0"
  xmlns:xupdate="http://www.xmldb.org/xupdate"
  xmlns:ftss="%s"
  xmlns:xlink="%s"
>
  <xupdate:remove select="/*/ftss:CreationParams"/>
  <xupdate:append select="/*" child='last()'>
    %s
  </xupdate:append>

</xupdate:modifications>
"""
