import cStringIO
import Ft.Xml.Xslt.Processor

from Ft.Rdf import Statement
from Ft.Server import RESERVED_NAMESPACE, FTSERVER_NAMESPACE
from Ft.Server.Common import ResourceTypes, Schema, AclConstants, XmlLib
from Ft.Server.Server import FtServerServerException, Error
from Ft.Server.Server.Drivers import FtssInputSource
from Ft.Xml import XUpdate
from Ft.Xml.XLink import XLINK_NAMESPACE
from Ft.Xml.XPath import Context
from Ft.Xml.Domlette import Print
from Ft.Rdf import OBJECT_TYPE_LITERAL, OBJECT_TYPE_RESOURCE
import Ft.Server.Server.Xslt

class ResourceMetaDataImp:

    """A MetaData object is a piece of all objects that contains information about its existence.
    """

    resourceType = ResourceTypes.ResourceType.META_DATA

    def __init__(self, path, driver, cache):

        self._path = path
        self._deleted = 0
        self._driver = driver
        self._cache = cache
        self._cache.registerResource(self._path, self)
        self._basePath = self.getBasePath()

    ###############################################
    #General interfaces for getting information about the object
    ###############################################
    def getPath(self):
        """
        Get the path of this resource
        """
        self._verifyTx()
        return self._path

    def getBasePath(self):
        """
        Get the path of this object as a base path
        """
        self._verifyTx
        return self._path.getParentPath()

    def getDisplayPath(self):
        """
        Get the path relative to the document root
        """
        self._verifyTx()
        return self._path.displayPath

    def getAbsolutePath(self):
        """
        Get the full path into the system
        """
        self._verifyTx()
        return self._path.absolutePath

    def getName(self):
        """
        Get the name of the resource.
        """
        return self._path.name


    def getCreationDate(self):
        """
        Get the date that this resource was created.
        The date will be an ISO 8601 string like u'2002-08-16T22:49:56Z'.
        """
        self._verifyTx()
        res = self._driver.getSystemModel().complete(self._path.absolutePath, Schema.CREATION_DATE, None)
        if not res:
            raise FtServerServerException(Error.UNKNOWN_PATH,
                                          path=self._path)
        return res[0].object

    def getLastModifiedDate(self):
        """
        Get the date that this resource was last modified.
        The date will be an ISO 8601 string like u'2002-08-16T22:49:56Z'.
        """
        self._verifyTx()
        res = self._driver.getSystemModel().complete(self._path.absolutePath, Schema.MODIFIED_DATE, None)
        if not res:
            raise FtServerServerException(Error.UNKNOWN_PATH,
                                          path=self._path)
        return res[0].object

    def getSize(self):
        """
        Get the size of the content, in bytes, as an integer.
        """
        self._verifyTx()

        res = self._driver.getSystemModel().complete(self._path.absolutePath, Schema.CONTENT_SIZE, None)
        if not res:
            raise FtServerServerException(Error.UNKNOWN_PATH,
                                          path=self._path)
        return int(res[0].object)

    def getResourceType(self):
        """
        Get the type of this resource.  Types are defined in Core.Constants.ResourceType
        """
        self._verifyTx()
        return self.resourceType

    def isResourceType(self, testType):
        """
        Determine if this resource is of a certain type
        """
        self._verifyTx()
        return ResourceTypes._IsResourceType(self.getResourceType(), testType)

    def getContent(self):
        """
        Get the string content of this resource
        """
        self._verifyTx()
        return self._driver.fetchResource(self._path)

    def getContentSlice(self, start=None, end=None):
        """
        Get a piece of our content.  Really only makes sense for large raw files
        """
        if start:
            if end:
                return self.getContent()[start:end]
            return self.getContent()[start:]
        elif end:
            return self.getContent()[:end]
        return self.getContent()

    def getMetaDataResource(self):
        """
        Get the string meta data of this resource
        """
        self._verifyTx()
        return self

    def getContentResource(self):
        """
        Get the resource the represents the content of this object
        """
        self._verifyTx()
        p = self._path.normalize(".;content",0)
        return self._fetchResource(p)


    def getImt(self):
        """
        Returns the Internet Media Type of the raw file resource
        """
        self._verifyTx()
        return 'text/xml'

    def getValidationInfo(self):
        """
        Get the validation information associated with this resource.
        """
        self._verifyTx()
        return None


    def getParent(self):
        """
        Get the parent container of this resource
        """
        self._verifyTx()

        parentPath = self._path.getParentPath()
        return self._fetchResource(parentPath)

    def getRoot(self):
        """
        Get a reference to the system repository
        """
        self._verifyTx()
        return self._cache.getResource(self._path._documentRoot)

    def getAliases(self):
        """
        Get a list of alias objects that reference this object
        """
        self._verifyTx()
        aliases = []
        for aRes in self._odsResource.aliases:
            uri = self._xa._buildUriFromOdsResource(aRes)
            aliases.append(self.fetchResource(uri,traverseAliases=0))
        return aliases

    def getExtFunctionsAndElements(self):
        """
        All resource get to use the builtin functions
        """
        return Ft.Server.Server.Xslt.ExtFunctions.copy(),Ft.Server.Server.Xslt.ExtElements.copy()
        


    #####################################
    #ACL Interfaces
    #####################################
    def addAcl(self, aclKey, aclIdent, allowed=1):
        """
        Add an acl identifier to the acl key
        """
        self._verifyTx()
        self._verifyAclIdent(aclIdent)
        acl = self._driver.getAcl(self._path)
        if not acl.has_key(aclKey):
            acl[aclKey] = {}
        if allowed:
            acl[aclKey][aclIdent] = AclConstants.ALLOWED
        else:
            acl[aclKey][aclIdent] = AclConstants.DENIED
        self._driver.setAcl(self._path,acl)

    def getAcl(self, aclKey=None):
        """
        Get a list of users and groups specified by the aclKey
        """
        self._verifyTx()
        return self._driver.getAcl(self._path, aclKey)

    def getAclIdentifiers(self):
        """
        Get a list of all acl identifiers for the current users
        """
        self._verifyTx()
        return self._driver.getAclIdentifiers()

    def getAclIdent(self):
        """
        Get the ACL identifier for the current user
        """
        self._verifyTx()
        return self._driver.getAclIdent()

    def inheritAcl(self, aclKey):
        """
        Remove the entire key from the ACL entry so that it is inherited again from its parent
        """
        self._verifyTx()
        acl = self._driver.getAcl(self._path)
        if not acl.has_key(aclKey):
            return
        del acl[aclKey]
        self._driver.setAcl(self._path,acl)


    def removeAcl(self, aclKey, aclIdent):
        """
        Remove the acl identifier from the acl key
        """
        self._verifyTx()
        self._verifyAclIdent(aclIdent)
        acl = self._driver.getAcl(self._path)
        if not acl.has_key(aclKey):
            return
        if not acl[aclKey].has_key(aclIdent):
            return
        del acl[aclKey][aclIdent]
        self._driver.setAcl(self._path,acl)


    def setAcl(self, aclKey, aclIdent, allowed=1):
        """
        Replace the aclKey with a single entry of aclIdent
        """
        self._verifyTx()
        self._verifyAclIdent(aclIdent)
        acl = self._driver.getAcl(self._path)
        if allowed:
            acl[aclKey] = {aclIdent:AclConstants.ALLOWED}
        else:
            acl[aclKey] = {aclIdent:AclConstants.DENIED}
        self._driver.setAcl(self._path,acl)

    def verifyAcl(self, aclKey, verifyTraverse=1):
        """
        Verify that the testAclIdent is in the acl list specified by the acl key for this object.
        """
        self._verifyTx()
        return self._driver.verifyAcl(self._path, aclKey, verifyTraverse)

    def _verifyAclIdent(self, aclIdent):
        if aclIdent in [AclConstants.WORLD_GROUP_NAME,
                        AclConstants.USERS_GROUP_NAME,
                        AclConstants.SUPER_USER_GROUP_NAME,
                        AclConstants.OWNER]:
            return 1
        if aclIdent in self.getRoot().getAllUserNames():
            return 1
        if aclIdent in self.getRoot().getAllGroupNames():
            return 1
        raise FtServerServerException(Error.INVALID_ACL_IDENT,
                                      aclIdent=aclIdent)

    def getOwner(self):
        """
        Get the owner of this resource
        """
        self._verifyTx()
        ownerName = self._driver.getSystemModel().complete(self.getAbsolutePath(),
                                                           Schema.OWNER,
                                                           None)[0].object
        return self.getRoot().fetchUserOrGroupByName(ownerName)

    def setOwner(self, newOwner):
        """
        Change the owner of this resource
        """
        self._verifyTx()
        self._driver.verifyAcl(self.getPath(), AclConstants.CHANGE_OWNER_ACCESS, 0)
        xu = SET_OWNER_XUPDATE%(FTSERVER_NAMESPACE,XLINK_NAMESPACE,newOwner.getName())
        xu = XmlLib.MakeString(xu)
        self._driver.xupdateMetaData(self._path, xu)


    ##########################################
    #XML Interfaces
    ##########################################

    def applyXslt(self, stylesheets, params=None, ignorePis=1,
                  extParams=None, extModules=None):
        """
        applies the specified stylesheets (with the specified transformation parameters)
        on the DOM representation of the raw file resource
        """
        extModules = extModules or []

        params = params or {}
        params[(RESERVED_NAMESPACE, 'display-path')] = self._path.displayPath
        params[(RESERVED_NAMESPACE, 'absolute-path')] = self._path.absolutePath
        if self._path.absolutePath == '/':
            params[(RESERVED_NAMESPACE, 'base-path')] = self._basePath.absolutePath
        else:
            params[(RESERVED_NAMESPACE, 'base-path')] = self._basePath.absolutePath + '/'
        extParams = extParams or {}
        extParams[(RESERVED_NAMESPACE, 'path')] = self._path.absolutePath

        #First map the stylesheets
        if type(stylesheets) not in [type([]),type(())]:
            stylesheets = (stylesheets,)

        p = Ft.Xml.Xslt.Processor.Processor()
        p.extensionParams = extParams

        # Register our general extension modules
        p.registerExtensionModules(['Ft.Server.Server.Xslt']+extModules)

        p._repository = self.getRoot()

        p.setDocumentReader(
            FtssInputSource.NonvalidatingReader
            )

        for s in stylesheets:
            if type(s) == type(''):
                #It is a URI
                isrc = FtssInputSource.FtssInputSourceFactory.fromUri(self.getAbsolutePath(),self._driver)
                isrc = isrc.resolve(s,'','Repo ApplyXlst')
                p.appendStylesheet(isrc)
            elif hasattr(s,'toStylesheet'):
                #A document reference
                p.appendStylesheetInstance(s.toStylesheet(self))
            elif hasattr(s,'asStylesheet'):
                #A reference to a document in the system
                p.appendStylesheetInstance(s.asStylesheet())

                #See if it has any ext functions
                f,e = s.getExtFunctionsAndElements()
                p.extFunctions.update(f)
                p.extElements.update(e)


        isrc = FtssInputSource.FtssInputSourceFactory.fromUri(self.getAbsolutePath(),self._driver)

        rt = p.run(isrc, ignorePis=ignorePis, topLevelParams=params)

        if extParams is not None and hasattr(p, 'extensionParams'):
            extParams.update(p.extensionParams)

        imt = p._lastOutputParams.mediaType
        return rt, imt

    def asDom(self,stripElements=None):
        """
        Return this object as a non-live DOM instance
        """
        self._verifyTx()
        return FtssInputSource.NonvalidatingReader.parseString(self.getContent(),
                                                               self.getAbsolutePath(),
                                                               self._driver,
                                                               stripElements = stripElements)
    def toXPathContext(self,nss=None):
        """
        Create an XPath Context with the src of this RawFile
        """
        self._verifyTx()
        return self._driver.getContext(self._path,nss)


    def xUpdate(self, updateSrc):
        """
        Allows XML content to be updated with the XUpdate protocol
        """
        xureader = XUpdate.Reader()

        isrc = FtssInputSource.FtssInputSourceFactory.fromString(
            updateSrc, self._path.absolutePath, self._driver
        )
        xu = xureader.fromSrc(isrc)
        processor = XUpdate.Processor()
        dom = self.asDom()
        processor.execute(dom, xu)

        st = cStringIO.StringIO()
        Print(dom,stream=st)
        self.setContent(st.getvalue())


    #####################################
    #Query Interfaces
    #####################################
    def hasResource(self,path):
        """
        Query if the system has a resource specified by the path
        """
        self._verifyTx()
        path = self._basePath.normalize(path,0)
        if self._driver.hasResource(path):
            return self._driver.getType(path)
        return 0


    #####################################
    #Retrieval Interfaces
    #####################################
    def fetchResource(self,path,traverseAliases = 1):
        """
        Fetch an resource from the system.  If traverseAliases is true and the resource being fetched
        is an alias, it will traverse the alias path until a non alias resource is found (or a
        BROKEN_ALIAS exception is raised).
        """
        self._verifyTx()
        path = self._basePath.normalize(path,traverseAliases)
        return self._fetchResource(path)


    ##########################################
    #Retrieval Interfaces
    ##########################################

    def _fetchResource(self,path):

        if path.displayPath == '/' and path.resourceType != ResourceTypes.RESOURCE_METADATA:
            return self.getRoot()

        obj = self._cache.getResource(path)
        if obj is not None:
            return obj

        self._driver.verifyAcl(path,AclConstants.READ_ACCESS,1)

        if path.resourceType == ResourceTypes.RESOURCE_METADATA:
            #Get the MetaData object
            return ResourceMetaDataImp(path,self._driver,self._cache)

        rt = self._driver.getType(path)

        if rt == ResourceTypes.ResourceType.RAW_FILE:
            import RawFileImp
            obj = RawFileImp.RawFileImp(path,self._driver,self._cache)
        elif rt == ResourceTypes.ResourceType.URI_REFERENCE_FILE:
            import UriReferenceFileImp
            obj = UriReferenceFileImp.UriReferenceFileImp(path,self._driver,self._cache)
        elif rt == ResourceTypes.ResourceType.XML_DOCUMENT:
            import XmlDocumentImp
            obj = XmlDocumentImp.XmlDocumentImp(path,self._driver,self._cache)
        elif rt == ResourceTypes.ResourceType.COMMAND:
            import CommandImp
            obj = CommandImp.CommandImp(path,self._driver,self._cache)
        elif rt == ResourceTypes.ResourceType.SERVER:
            import ServerImp
            obj = ServerImp.ServerImp(path,self._driver,self._cache)
        elif rt == ResourceTypes.ResourceType.XSLT_DOCUMENT:
            import XsltDocumentImp
            obj = XsltDocumentImp.XsltDocumentImp(path,self._driver,self._cache)
        elif rt == ResourceTypes.ResourceType.RDF_DOCUMENT:
            import RdfDocumentImp
            obj = RdfDocumentImp.RdfDocumentImp(path,self._driver,self._cache)
        elif rt == ResourceTypes.ResourceType.CONTAINER:
            import ContainerImp
            obj = ContainerImp.ContainerImp(path,self._driver,self._cache)
        elif rt == ResourceTypes.ResourceType.GROUP:
            import GroupImp
            obj = GroupImp.GroupImp(path,self._driver,self._cache)
        elif rt == ResourceTypes.ResourceType.USER:
            import UserImp
            obj = UserImp.UserImp(path,self._driver,self._cache)
        elif rt == ResourceTypes.ResourceType.XPATH_DOCUMENT_DEFINITION:
            import XPathDocumentDefinitionImp
            obj = XPathDocumentDefinitionImp.XPathDocumentDefinitionImp(path,self._driver,self._cache)
        elif rt == ResourceTypes.ResourceType.XSLT_DOCUMENT_DEFINITION:
            import XsltDocumentDefinitionImp
            obj = XsltDocumentDefinitionImp.XsltDocumentDefinitionImp(path,self._driver,self._cache)
        elif rt == ResourceTypes.ResourceType.ALIAS:
            import AliasImp
            obj = AliasImp.AliasImp(path,self._driver,self._cache)
        elif rt == ResourceTypes.ResourceType.SCHEMATRON_DOCUMENT:
            import SchematronDocumentImp
            obj = SchematronDocumentImp.SchematronDocumentImp(path,self._driver,self._cache)
        else:
            raise FtServerServerException(Error.UNKNOWN_RESOURCE_TYPE,type=rt)
        return obj


    ##########################################
    #Mutation Interfaces
    ##########################################
    def setContent(self,src):
        """
        Set the string content of this resource
        """
        self._verifyTx()
        self._driver.updateResourceContent(self._path,src)


    def delete(self):
        """
        Delete this object from the repository
        """
        self._verifyTx()
        self._driver.verifyDeleteResource(self._path)
        self._delete()
        self._driver.deleteResource(self._path)

        self._cache.unregisterResource(self._path)

        #Just incase
        mdPath = self._path.normalize('.;metadata',0)
        self._cache.unregisterResource(mdPath)
        conPath = self._path.normalize('.;content',0)
        self._cache.unregisterResource(conPath)

        self._deleted = 1



    ###################################
    #Creation Interfaces
    ###################################
    def addAlias(self,path,docDef = None):
        """
        Adds an alias for this resource to the repository.
        """
        self._verifyTx()
        path = self._basePath.normalize(path,1)
        if docDef:
            docDefPath = self._basePath.normalize(docDef,1)
            if self._driver.getType(docDefPath) not in ResourceTypes.DOCUMENT_DEFINITIONS:
                raise FtServerServerException(Error.INVALID_PATH,
                                              path=docDefPath.displayPath,
                                              type='Document Definition')
        else:
            docDefPath = None

        import AliasImp
        md,content = AliasImp.NewAliasXml(self._driver,
                                          self._path.absolutePath,
                                          path.name,
                                          {},
                                          self._driver.getAclIdent(),
                                          docDefPath and docDefPath.absolutePath or None)

        self._driver.createResource(path,md,content)
        return self._fetchResource(path)



    ######################################
    #Delete Interface
    ######################################
    def removeAlias(self,path):
        """
        Removes the specified alias for this resource
        """
        self._verifyTx()
        path = self._basePath.normalize(path,0)
        res = self._fetchResource(path)
        res.delete()
        return


    ########################################
    #RDF Interfaces
    ########################################

    def getModel(self):
        """
        Retrieves the common model associated w/ the repository
        """
        self._verifyTx()
        return self._driver.getModel(self._path)

    def addLiteralProperty(self, predicate, object, scope='', statementUri=''):
        """
        Adds a statement about this repository object, with a literal property, to
        the model
        """
        stmt = Statement.Statement(self.getAbsolutePath(),
                                   predicate,
                                   object,
                                   scope = scope,
                                   statementUri=statementUri,
                                   objectType=OBJECT_TYPE_LITERAL)
        self._driver.getModel(self._path).add([stmt])
        return stmt

    def addResourceProperty(self, predicate, object, scope='', statementUri=''):
        """
        Adds a statement about this repository object, with a resource
        property, to the model
        """
        stmt = Statement.Statement(self.getAbsolutePath(),
                                   predicate,
                                   object.getUri(),
                                   scope=scope,
                                   statementUri=statementUri,
                                   objectType=OBJECT_TYPE_RESOURCE)
        self._driver.getModel(self._path).add([stmt])
        return stmt

    def removeResourceProperty(self, predicate, object):
        """
        Remove a statement about this repository object, with a resource preoperty, from
        the model
        """
        self._driver.getModel(self._path).removePattern(self.getAbsolutePath(),predicate,object.getUri())

    def removeLiteralProperty(self, predicate, object):
        """
        Removes a statement about this repository object, with a literal property, from
        the model
        """
        self._driver.getModel(self._path).removePattern(self.getAbsolutePath(),predicate,object)

    def getStatementsAbout(self):
        """
        Returns all statements about this object in the model
        """
        return self._driver.getModel(self._path).complete(self.getPath(), None, None)

    def getStatementsGeneratedBy(self):
        """
        Returns all statements in the model generated by this object
        """
        return self._driver.getModel(self._path).complete(None, None, None, scope=self.getPath())


    ########################################
    #Temporary file Interfaces
    ########################################
    def markTemporary(self, timeToLive):
        """
        Mark this resource as temporary.  Temporary objects will be removed from the repository when thier
        timeToLive has expired and purgeTemporaryUris is called the repository.
        Time to live is seconds from now.
        """
        self.getRoot().setPathTemporary(self.getAbsolutePath(), timeToLive)


    ##########################################
    #String interfaces
    ##########################################
    def __str__(self):
        return "<%s at %x path='%s'>" % (ResourceTypes.g_resourceTypeMap[self.getResourceType()],
                                         id(self),
                                         self.getPath().absolutePath)


    ######################################
    #Internal Interfaces
    ######################################

    def _verifyTx(self):
        self._driver._verifyTx()
        if self._deleted:
            raise FtServerServerException(Error.OBJECT_DELETED, path=self._path)

    def _delete(self):
        """Some don't define this"""
        pass


SET_OWNER_XUPDATE="""<xupdate:modifications
  version="1.0"
  xmlns:xupdate="http://www.xmldb.org/xupdate"
  xmlns:ftss="%s"
  xmlns:xlink="%s"
>
<xupdate:update select="/ftss:MetaData/ftss:Owner">%s</xupdate:update>
</xupdate:modifications>"""



UPDATE_TIME_TO_LIVE_XUPDATE="""<?xml version="1.0"?>
<xupdate:modifications
  version="1.0"
  xmlns:xupdate="http://www.xmldb.org/xupdate"
  xmlns:ftss="%s"
  xmlns:xlink="%s"
>
<xupdate:update select="/ftss:MetaData/ftss:TimeToLive">%s</xupdate:update>
</xupdate:modifications>
"""

SET_TIME_TO_LIVE_XUPDATE="""<?xml version="1.0"?>
<xupdate:modifications
  version="1.0"
  xmlns:xupdate="http://www.xmldb.org/xupdate"
  xmlns:ftss="%s"
  xmlns:xlink="%s"
>
<xupdate:append select="/ftss:MetaData"><ftss:TimeToLive>%s</ftss:TimeToLive></xupdate:append>
</xupdate:modifications>
"""
