########################################################################
#
# File Name:            Handler.py
#
"""
SAX event handler class for minidom readers

WWW: http://4suite.org        e-mail: support@4suite.org

Copyright (c) 1999-2001 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

from types import TupleType
from xml.sax import xmlreader, handler

from Ft.Xml import XML_NAMESPACE, EMPTY_NAMESPACE, XMLNS_NAMESPACE
from Ft.Xml import SplitQName

from Ft.Xml import ReaderException

from Ft.Xml.XInclude import XINCLUDE_NAMESPACE, XIncludeException

from Strip import XmlStrStrip
from Nodes import Node

class Handler:

    def initHandler(self, inputSource):
        self.inputSource = inputSource

        # Init our root node
        self._rootNode = self._ownerDoc = self.Document()
        self._rootNode.refUri = inputSource.uri

        # Setup the stack which keeps track of the nesting of DOM nodes.
        self._nodeStack = [self._rootNode]
        self._namespaces = [{'xml': XML_NAMESPACE,'xmlns':XMLNS_NAMESPACE}]
        self._currText = ''

        self._preserveStateStack = [1]  #initialize to preserve
        self._stripElements = inputSource.stripElements

        # Setup our include depth
        self._includeDepth = 0

        # Setup the inline indexing
        self._currIndex = 1
        self._rootNode.docIndex = 1

        self._xmlBaseStack = [inputSource.uri]
        self._rootNode.xmlBase = inputSource.uri

        # For XIncludes
        self._visited_hrefs = []
        self._ignore_top_level_whitespace = [1]
        return

    def getRootNode(self):
        self._completeTextNode()
        return self._rootNode

    ### ContentHandler Interface ###

    def startElement(self, name, attribs):

        self._completeTextNode()

        self._ignore_top_level_whitespace.append(0)

        # Create a copy of our parents namespaces
        self._namespaces.append(self._namespaces[-1].copy())

        #Read out attributes
        for curr_attrib_key in attribs.keys():
            if type(curr_attrib_key) is TupleType:
                local = curr_attrib_key[1]
                prefix = None
                if curr_attrib_key[0]:
                    raise "Namespaces in validating docs not supported"
            else:
                (prefix, local) = SplitQName(curr_attrib_key)
            if not local:
                raise ReaderException(ReaderException.INVALID_XMLNS, curr_attrib_key)
            if prefix == 'xmlns':
                self._namespaces[-1][local] = attribs[curr_attrib_key]
            elif prefix is None and local == 'xmlns':
                self._namespaces[-1][None] = attribs[curr_attrib_key]

        if type(name) is TupleType:
            local = name[1]
            prefix = None
            if name[0]:
                raise "Namespaces in validating docs not supported"
        else:
            (prefix, local) = SplitQName(name)

        #if prefix and not self._namespaces[-1].has_key(prefix):
        #    raise ReaderException(ReaderException.XMLNS_UNKNOWN_PREFIX, prefix)

        if self._namespaces[-1].has_key(prefix):
            namespace = self._namespaces[-1][prefix]
        else:
            namespace = EMPTY_NAMESPACE

        attrs = {}
        qnames = {}
        for curr_attrib_key in attribs.keys():
            if type(curr_attrib_key) is TupleType:
                a_local = curr_attrib_key[1]
                a_prefix = None
                if curr_attrib_key[0]:
                    raise "Namespaces in validating docs not supported"
            else:
                (a_prefix, a_local) = SplitQName(curr_attrib_key)
            if a_prefix and self._namespaces[-1].has_key(a_prefix):
                ns = self._namespaces[-1][a_prefix]
            else:
                ns = EMPTY_NAMESPACE
            attrs[(ns, a_local)] = attribs[curr_attrib_key]
            qnames[(ns, a_local)] = curr_attrib_key

        nsattribs = xmlreader.AttributesNSImpl(attrs, qnames)

        #Adjust local variables
        (name, qname) = ((namespace or None, local), name)

        #See if we need to handle XInclude

        if self.inputSource.processIncludes and name == (XINCLUDE_NAMESPACE, 'include'):

            if self._includeDepth:
                self._includeDepth = self._includeDepth + 1
            else:
                #Looks like it is a GO!!
                href = nsattribs.get((EMPTY_NAMESPACE, 'href'))
                if not href:
                    raise XIncludeException(XIncludeException.XINCLUDE_MISSING_HREF)
                source = self.inputSource.resolve(href,'','XINCLUDE')
                if source.uri in self._visited_hrefs:
                    raise XIncludeException(
                        XIncludeException.CIRCULAR_INCLUDE_ERROR, href
                        )

                parse = nsattribs.get((EMPTY_NAMESPACE, 'parse'))
                if not parse or parse == 'xml':

                    self._visited_hrefs.append(source.uri)
                    self._ignore_top_level_whitespace.append(1)

                    orig_parser = self.parser
                    orig_source = self.inputSource

                    self.prepareParser()
                    self.inputSource = source

                    self.parse(source)

                    self.parser = orig_parser
                    self.inputSource = orig_source

                    del self._visited_hrefs[-1]
                    del self._ignore_top_level_whitespace[-1]
                else:
                    self._currText = self._currText + source.stream.read()
                source.stream.close()
                self._includeDepth = 1
        else:  #end of the check for XInclude processing

            #Create the new Element
            new_element = self.Element(qname, namespace, prefix, local)
            new_element.ownerDocument = self._ownerDoc
            self._docIndex(new_element)


            nextBase = self._xmlBaseStack[-1]

            for attr_qname in nsattribs.getQNames():

                attr_ns = nsattribs.getNameByQName(attr_qname)[0]
                (attr_prefix, attr_local) = SplitQName(attr_qname)

                if attr_prefix is None and attr_local == 'xmlns':
                    #This is a default NS
                    attr_ns = XMLNS_NAMESPACE
                    attr_key = (attr_ns, None)
                else:
                    attr_key = (attr_ns, attr_local)

                attr = self.Attr(attr_qname, attr_ns, attr_local, attr_prefix)
                attr.value = nsattribs.getValueByQName(attr_qname)
                attr.ownerElement = new_element
                attr.ownerDocument = self._ownerDoc
                attr.xmlBase = self._xmlBaseStack[-1]
                self._docIndex(attr)
                new_element._attrsNS[attr_key] = attr

                #Look for a change in xml:base
                if (XML_NAMESPACE,'base') == attr_key:
                    nextBase = attr.value

            new_element.xmlBase = nextBase
            self._xmlBaseStack.append(nextBase)

            new_pstate = self._preserveStateStack[-1]
            for (uri, local, strip) in self._stripElements:
                if (uri, local) in [(new_element.namespaceURI,
                                     new_element.localName), (EMPTY_NAMESPACE, '*'),
                                    (new_element.namespaceURI, '*')
                                    ]:
                    new_pstate = not strip
                    break
            self._preserveStateStack.append(new_pstate)

            self._nodeStack.append(new_element)
        return

    def endElement(self, name):

        del self._namespaces[-1]

        if self._includeDepth:
            self._includeDepth = self._includeDepth - 1
            del self._ignore_top_level_whitespace[-1]
            return

        self._completeTextNode()
        del self._ignore_top_level_whitespace[-1]

        new_element = self._nodeStack[-1]

        del self._preserveStateStack[-1]

        del self._xmlBaseStack[-1]

        del self._nodeStack[-1]
        self._nodeStack[-1].appendChild(new_element)
        return

    def characters(self, data):
        if self._includeDepth: return
        self._currText = self._currText + data
        return

    def processingInstruction(self, target, data):
        if self._includeDepth: return
        self._completeTextNode()
        pi = self.ProcessingInstruction(target,data)
        pi.ownerDocument = self._ownerDoc
        pi.xmlBase = self._xmlBaseStack[-1]
        self._docIndex(pi)
        self._nodeStack[-1].appendChild(pi)
        return

    ### LexicalHandler Interface ###

    def comment(self, data):
        if self._includeDepth: return
        self._completeTextNode()
        comment = self.Comment(data)
        comment.ownerDocument = self._ownerDoc
        comment.xmlBase = self._xmlBaseStack[-1]
        self._docIndex(comment)
        self._nodeStack[-1].appendChild(comment)
        return

    ### Internal Functions ###

    def _completeTextNode(self):
        if not self._currText:
            return
    
        data = self._currText
        self._currText = u''

        # Ignore any top-level whitespace
        whitespace_only = not XmlStrStrip(data)
        if whitespace_only and self._ignore_top_level_whitespace[-1]:
            return

        if self._preserveStateStack[-1] or not whitespace_only:
            # Create a new text node
            text = self.Text(data)
            text.ownerDocument = self._ownerDoc
            text.xmlBase = self._xmlBaseStack[-1]
            self._docIndex(text)
            top_node = self._nodeStack[-1]
            top_node.appendChild(text)
        return

    def _docIndex(self, node):
        self._currIndex = self._currIndex + 1
        node.docIndex = self._currIndex
        return

#Set the default nodes
import Nodes
Handler.Document = Nodes.Document
Handler.ProcessingInstruction = Nodes.ProcessingInstruction
Handler.Element = Nodes.Element
Handler.Attr = Nodes.Attr
Handler.Text = Nodes.Text
Handler.Comment = Nodes.Comment

