########################################################################
#
# File Name:            InputSource.py
#
#
"""
An Input Source is an encapsulation of a source of content.
It includes a URI/system ID and processing parameters used by the parsers.

WWW: http://4suite.org        e-mail: support@4suite.org

Copyright 2002 Fourthought Inc, USA.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

from Ft.Lib import Uri, Uuid
import cStringIO, types, warnings


class InputSource:
    """
    An Input Source is an encapsulation of a source of content.
    It includes a URI/system ID and processing parameters used by the parsers.

    It is designed to be overridden as applications need different functionality
    from sources.
    """
    def __init__(self, stream, uri=None, processIncludes=1, stripElements=None,
                 factory=None):
        """
        stream - the stream associated with this input source
        uri - the URI or system ID of the input source
        processIncludes=1 - Whether or not XIncludes should be expanded
        stripElements - Space stripping rules
        factory - The factory that created this
        """
        if uri:
            self.uri = uri
        else:
            self.uri = 'urn:uuid:'+Uuid.UuidAsString(Uuid.GenerateUuid())
        self.stream = stream
        self.processIncludes = processIncludes
        self.stripElements = stripElements or []
        self.factory = factory
        frag_split = self.uri.split("#")
        self.fragment = ""
        if len(frag_split) == 2:
            self.fragment = frag_split[1]
        return

    def resolve(self, sysid, pubid=None, hint=None):
        """
        Resolve a system id and or a public id into a new input source.
        The hint parameter is used to give a hint as to what the resolution will be used for.
        The default behaviour ignores the pubid and hint.
        """
        uri = self._normalize(sysid)
        stream = self._openStream(uri)
        return self._clone(stream, uri)

    def getUriResolver(self):
        """
        This method will return a URI resolver that is used by this input source to normalize URIs.
        This is the public method to use if just URI resolution is needed
        """
        return Uri.BASIC_RESOLVER

    #Stream protocol methods
    def read(self, bufsiz=-1):
        return self.stream.read(bufsiz)

    def close(self):
        return self.stream.close()

    #Helper Methods
    def _normalize(self, uri):
        """
        Normalize a uri with the ref URI of this input source.
        The default implementation will just use the default URI resolver

        If your input source is working with non-standard or not supported
        URIs, then you will need to override this or the getUriResolver method.
        """
        return self.getUriResolver().normalize(uri, self.uri)

    def _openStream(self, uri):
        """
        Open the given (already normalized) uri into a stream.
        Default bahviour uses the default resolver open on the URI

        If your custom InputSource is required to open URIs that are not supported
        nativly by this InputSource (example repository objects, or objects from a database),
        then you should override this method and do what ever it takes to resolve the URI into
        a readable stream.
        """
        return self.getUriResolver().resolve(uri)

    def _clone(self, stream, uri=None):
        """
        Clone this input source.  Create a new instance of with
        the known params

        If, you derived InputSource requires additional state information then you need
        to override how it is cloned and pickled.
        """
        if uri == None: uri = self.uri
        return self.__class__(stream, uri,
                              processIncludes=self.processIncludes,
                              stripElements=self.stripElements,
                              factory=self.factory)

    #Pickle routines.  you need to be able to pickle an input source
    #We cannot pickle a stream though.
    def __getstate__(self):
        state = self.__dict__.copy()
        state['stream'] = None
        return state


class InputSourceFactory:

    def __init__(self, inputSourceClass=None):
        self._klass = inputSourceClass or InputSource
        return

    def fromUri(self, uri, *v_args, **kw_args):
        """
        parse the file retrieved from the given URI
        """
        kw_args['factory'] = self
        src = apply(self.fromStream,(None, uri) + v_args,kw_args)
        stream = src._openStream(src.uri)
        return src._clone(stream, src.uri)

    def fromString(self, st, uri=None, *v_args, **kw_args):
        """
        parse the given string.  uri is the URI or system ID
        to use for the entity of the string
        """
        if not uri:
            warnings.warn("Creation of InputSource without an URI",
                          RuntimeWarning, 2)
        if not isinstance(st, types.StringType):
            raise ValueError("String must be a of type string, not %s" % type(st))
        kw_args['factory'] = self
        stream = cStringIO.StringIO(st)
        return apply(self.fromStream, (stream,uri)+v_args, kw_args)

    def fromStream(self, stream, uri=None, *v_args, **kw_args):
        """
        parse the content of the given stream.  uri is the URI or system ID
        to use for the entity of the string
        """
        if not uri:
            warnings.warn("Creation of InputSource without an URI",
                          RuntimeWarning, 2)
        kw_args['factory'] = self
        return apply(self._klass, (stream,uri)+v_args, kw_args)


DefaultFactory = InputSourceFactory()

