########################################################################
#
# File Name:   BuiltInExtFunctions.py
#
# Docs:        http://docs.4suite.org/XPath/BuiltInExtFunctions.py.html
#
"""
4XPath-specific Extension functions
WWW: http://4suite.org/XSLT        e-mail: support@4suite.org

Copyright (c) 2000-2001 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import os, re, urllib, codecs, Ft
import whrandom
from xml.dom import Node
from Ft.Lib import boolean
from Ft.Lib import Uri
from Ft.Xml.XPath import Conversions, FT_EXT_NAMESPACE, FT_OLD_EXT_NAMESPACE


def Version(context):
    from Ft.__init__ import __version__
    return __version__


def NodeSet(context, rtf):
    """Convert a result-tree fragment to a node-set"""
    if type(rtf) == type([]):
        return rtf
    if hasattr(rtf,'nodeType') and rtf.nodeType == Node.DOCUMENT_NODE:
        node_set = list(rtf.childNodes)
    else:
        node_set = [rtf]
    return node_set


def Match(context, pattern, arg=None):
    """Do a regular expression match against the argument"""
    if not arg:
        arg = context.node
    arg = Conversions.StringValue(arg)
    return re.match(pattern, arg) and boolean.true or boolean.false


def ResolvePath(context, base, rel):
    """Resolve a path against a base.  Similar to f:resolve-url, but allows the base to be just a path, not necessarily a full URL"""
    base = Conversions.StringValue(base)
    rel = Conversions.StringValue(rel)
    import urllib
    return urllib.basejoin(base, rel)


def ResolveUrl(context, base, rel):
    """Return the relative URL ref resolved against the given base"""
    base = Conversions.StringValue(base)
    rel = Conversions.StringValue(rel)
    return Uri.BASIC_RESOLVER.normalize(rel, base)


def SharePath(context):   
    """ Returns the system-dependent path to Ft.Share """
    return Ft.SHARE_DIR


def Replace(context, old, new, arg=None):
    """Do a global search and replace of the string contents"""
    if not arg:
        arg = context.node
    arg = Conversions.StringValue(arg)
    old = Conversions.StringValue(old)
    new = Conversions.StringValue(new)
    return arg.replace(old, new)


def EscapeUrl(context, url):
    "Escape illegal characters in a URL"
    return urllib.quote(Conversions.StringValue(url))


def EscapeXml(context, text):
    """
    Escape illegal characters in XML text
    Should only really be relevant in text output mode
    """
    #FIXME: This should really also take illegal chars, such as control chars,
    #And convert them to char entities.  See if we can borrow someone's code
    #That does this rather than reinventing the wheel.
    from xml.sax.saxutils import escape
    return escape(Conversions.StringValue(text))

def BaseUri(context, arg=None):
    """Get the base URI of the argument"""
    #FIXME: Arg must be a node set
    if not arg:
        arg = [context.node]
    if hasattr(arg[0],'baseUri'):
        return arg[0].baseUri
    elif hasattr(arg[0],'refUri'):
        return arg[0].refUri
    return ""

def IsoTime(context):
    """
    Get the current date/time in ISO-8601 format.
    If all you want is the date, or all you want is the time,
    just use
    substring-before(isotime(), 'T')
    or
    substring-after(isotime(), 'T')
    """
    #try:
    #    import DateTime
    #    d = DateTime.now()
    #    iso_time = DateTime.ISO.str(d)
    #except ImportError:
    import time
    #(Y, M, D, h, m, s, d, j, dst) = time.localtime()
    t_tuple = time.localtime()
    iso_time = "%04i-%02i-%02iT%02i:%02i:%02i"%(t_tuple[:6])
    return iso_time

def Evaluate(context, expr):
    from Ft.Xml.XPath import parser
    return parser.new().parse(Conversions.StringValue(expr)).evaluate(context)


def GenerateUuid(context):
    from Ft.Lib import Uuid
    rt = Uuid.UuidAsString(Uuid.GenerateUuid())
    return rt


def Random(context,max=None,forceInt=0):
    if max:
        max = Conversions.NumberValue(max)
    else:
        max = 1
    rt = whrandom.random()*max
    if forceInt:
        rt = int(rt)
    return rt

##
## distinct, split, range if_function and find
## were contributed by Lars Marius Garshol.
## Their namespace URI was originally 'http://garshol.priv.no/symbolic/'
## but has been changed to the 4Suite.org NSRef
## so users don't have to declare yet another namespace.
##

def distinct(context, nodeset):
    if type(nodeset) != type([]):
        raise Exception("'distinct' parameter must be of type node-set!")

    nodes = {}
    for node in nodeset:
        nodes[Conversions.StringValue(node)] = node

    return nodes.values()

def split(context, arg, delim=None):

    doc = None
    if hasattr(context.node,'ownerDocument'):
        doc = context.node.ownerDocument or context.node
    else:
        doc = context.node
        while doc.parentNode:
            doc = doc.parentNode

    df = doc.createDocumentFragment()
    for token in Conversions.StringValue(arg).split(delim):
        df.appendChild(doc.createTextNode(token))

    return df.childNodes

def join(context, nodeset, delim=''):
    comps = map(lambda x: Conversions.StringValue(x), nodeset)
    if delim:
        return delim.join(comps)
    else:
        return ''.join(comps)

def EndsWith(context, outer, inner):
    """Function: <string> ends-with(<string>, <string>)"""
    outer = Conversions.StringValue(outer)
    inner = Conversions.StringValue(inner)
    return outer[-1*len(inner):] == inner and boolean.true or boolean.false


def range(context, lo, hi):
    doc = context.node
    while doc.parentNode:
        doc = doc.parentNode

    lo = Conversions.NumberValue(lo)
    hi = Conversions.NumberValue(hi)

    nodeset = []
    for number in xrange(lo, hi):
        nodeset.append(doc.createTextNode(str(number)))

    return nodeset

def if_function(context, cond, v1, v2=None):
    if Conversions.BooleanValue(cond):
        return v1
    elif v2 is None:
        return []
    else:
        return v2

def find(context, outer, inner):
    return Conversions.StringValue(outer).find(Conversions.StringValue(inner))

def DocAsString(context, object, nodeSet=None):
    result = []
    sheet = context.processor.stylesheet
    baseUri = context.currentInstruction.baseUri
    if nodeSet:
        baseUri = getattr(nodeSet[0].ownerDocument or nodeSet[0],
                          'refUri', baseUri)

    if nodeSet is None:
        if type(object) == type([]):
            for curr_node in object:
                result = result + Document(
                    context, Conversions.StringValue(curr_node),
                    [curr_node]
                    )
        else:
            try:
                # when the call is document(''), we want just the XML doc containing the
                # document('') call, not the entire stylesheet.
                if object == '':
                    uri = baseUri
                else:
                    #FIXME: Discard fragments before checking for dupes
                    uri = Uri.BASIC_RESOLVER.normalize(
                        Conversions.StringValue(object),
                        baseUri
                        )
                if object == '' and sheet.root.sources.has_key(baseUri):
                    result = sheet.root.sources[baseUri]
                else:
                    #Assume default codec (UTF-8?) for now
                    result =codecs.open(uri).read()
            except IOError:
                pass
    elif type(nodeSet) == type([]):
        if type(object) == type([]):
            for curr_node in object:
                result = result + Document(
                    context,
                    Conversions.StringValue(curr_node),
                    nodeSet
                    )
        else:
            try:
                #FIXME: Discard fragments before checking for dupes
                uri = Uri.BASIC_RESOLVER.normalize(
                    Conversions.StringValue(object),
                    baseUri
                    )
                #Assume default codec (UTF-8?) for now
                result =codecs.open(uri).read()
            except IOError:
                pass
    return result

def System(context, command):
    """
    Execute a command in an operating systems shell
    Caution: using this function could be a security hazard
    """
    command = Conversions.StringValue(command)
    result = os.system(command)
    return result

def Execv(context, command, *args):
    """
    Execute a command in an operating systems shell, passing in the
    command line arguments separately
    Caution: using this function could be a security hazard
    """
    command = Conversions.StringValue(command)
    result = os.execv(command, args)
    return result

def EnvVar(context, var):
    """
    Look up a variable in the OS environment
    Returns a string, either the environment variable value or an empty string if there is no such variable
    Caution: using this function could be a security hazard
    """
    var = Conversions.StringValue(var)
    result = os.environ.get(var, "")
    return result


def ParseXml(context, src, parameters=None):
    #baseUri - is the optionsl URI to be used for the document in the parsing.
    #          it is required if the string uses external entities or XInclude
    """
    f:xml-parse() parses a given string as XML and returns a node set whose
    sole item is the resulting parsed document root node.

    src    - converted to string value and parsed
    params - the name of a parameter set for the operation
             This argument is ignored for now, meaning the encoding must be UTF-8 and the XML must not use relative URIs in entities or XInclude

    An example:
    
    <xsl:stylesheet
      xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
      xmlns:f="http://xmlns.4suite.org/ext"
      version="1.0"
    >
      <xsl:variable name="doc" select="'&lt;spam xmlns:x=&quot;http://spam.com&quot;>eggs&lt;monty>python&lt;/monty>&lt;/spam>'"/>

      <xsl:template match="/">
      <xsl:value-of select="f:parse-xml($doc)/spam/monty"/>
      </xsl:template>
    </xsl:stylesheet>

    Run against any XML source should yield:

    python
    
    """
    from Ft.Xml import Domlette
    src = Conversions.StringValue(src).encode("utf-8")
    #Using FT_EXT_NAMESPACE as default URI is just a hack to avoid warnings
    #doc = Domlette.NonvalidatingReader.parseString(src, baseUri or FT_EXT_NAMESPACE)
    doc = Domlette.NonvalidatingReader.parseString(src, FT_EXT_NAMESPACE)
    return [doc]

def ParseDate(context, date, format=None):
    """
    This function is similar to exslt:parse-date
    http://exslt.org/date/functions/parse-date/index.html
    Except that it uses Python rather than Java conventions
    for the date formatting.
    """
    import time
    date = Conversions.StringValue(date)
    format = Conversions.StringValue(format)
    time_tuple = time.strptime(format)
    print time_tuple
    #perhaps add some variants for missing time tuple values?
    str_time = time.strftime("%Y-%m-%dT%H:%M:%S", time_tuple)
    return unicode(str_time)
    

ExtNamespaces = {
    FT_EXT_NAMESPACE : 'f',
    }

ExtFunctions = {
    (FT_EXT_NAMESPACE, 'node-set'): NodeSet,
    (FT_EXT_NAMESPACE, 'match'): Match,
    (FT_EXT_NAMESPACE, 'base-uri'): BaseUri,
    (FT_EXT_NAMESPACE, 'escape-url'): EscapeUrl,
    (FT_EXT_NAMESPACE, 'iso-time'): IsoTime,
    (FT_EXT_NAMESPACE, 'evaluate'): Evaluate,
    (FT_EXT_NAMESPACE, 'distinct'): distinct,
    (FT_EXT_NAMESPACE, 'split'): split,
    (FT_EXT_NAMESPACE, 'join'): join,
    (FT_EXT_NAMESPACE, 'ends-with'): EndsWith,
    (FT_EXT_NAMESPACE, 'range'): range,
    (FT_EXT_NAMESPACE, 'if'): if_function,
    (FT_EXT_NAMESPACE, 'find'): find,
    (FT_EXT_NAMESPACE, 'version'): Version,
    (FT_EXT_NAMESPACE, 'generate-uuid'): GenerateUuid,
    (FT_EXT_NAMESPACE, 'replace'): Replace,
    (FT_EXT_NAMESPACE, 'escape-xml'): EscapeXml,
    (FT_EXT_NAMESPACE, 'random'): Random,
    (FT_EXT_NAMESPACE, 'resolve-url'): ResolveUrl,
    (FT_EXT_NAMESPACE, 'resolve-path'): ResolvePath,
    (FT_EXT_NAMESPACE, 'doc-as-string'): DocAsString,
    (FT_EXT_NAMESPACE, 'share-path'): SharePath,
    (FT_EXT_NAMESPACE, 'system'): System,
    (FT_EXT_NAMESPACE, 'env-var'): EnvVar,
    (FT_EXT_NAMESPACE, 'execv'): Execv,
    (FT_EXT_NAMESPACE, 'parse-xml') : ParseXml,
    (FT_EXT_NAMESPACE, 'parse-date'): ParseDate,
    }

import MathFunctions

ExtFunctions.update(MathFunctions.ExtFunctions)

