########################################################################
#
# File Name:            RtfWriter.py
#
# Documentation:        http://docs.4suite.org/4XSLT/RtfWriter.py.html
#

"""
A special, simple writer for capturing result-tree fragments
WWW: http://4suite.org/4XSLT        e-mail: support@4suite.org

Copyright (c) 1999-2001 Fourthought Inc., USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import os
import NullWriter, OutputParameters
from Ft.Xml.Xslt import XSL_NAMESPACE
from Ft.Xml.XPath import Util
from Ft.Xml import SplitQName, XMLNS_NAMESPACE, EMPTY_NAMESPACE
from xml.dom import Node

class RtfWriter(NullWriter.NullWriter):
    def __init__(self, outputParams=None):
        self._root = Document()
        self._curr_node = self._root
        self._docIndex = 0
        self._root.docIndex = 0
        self._outputParams = outputParams or OutputParameters.OutputParameters()
        return

    def getResult(self):
        return self._root

    def startElement(self, name, namespace=EMPTY_NAMESPACE, extraNss=None):
        extraNss = extraNss or {}
        prefix, localName = SplitQName(name)
        element = Element(name, namespace, prefix, localName)
        element.docIndex = self._docIndex
        self._docIndex = self._docIndex + 1
        self._curr_node.appendChild(element)
        for (prefix, uri) in extraNss.items():
            if prefix:
                nodeName = u'xmlns:'+prefix
            elif not uri:
                # no prefix and no uri indicates default namespace NOT set
                continue
            else:
                nodeName = u'xmlns'
            attr = Attr(nodeName, XMLNS_NAMESPACE, prefix, u'xmlns', uri)
            attr.docIndex = self._docIndex
            self._docIndex = self._docIndex + 1
            element.setAttributeNodeNS(attr)
        self._curr_node = element
        return

    def endElement(self, name):
        parent = self._curr_node.parentNode
        
        parent.stringValue = parent.stringValue + self._curr_node.stringValue
        self._curr_node = parent
        return

    def attribute(self, name, value, namespace=EMPTY_NAMESPACE):
        if self._curr_node.attributes is not None:
            # Only add to elements
            prefix, localName = SplitQName(name)
            attr = Attr(name, namespace, prefix, localName, value)
            attr.docIndex = self._docIndex
            self._docIndex = self._docIndex + 1
            self._curr_node.setAttributeNodeNS(attr)
        return

    def text(self, data, escapeOutput=1, asis=0):
        #FIXME: Wrong.  This should be normalized!
        text = Text(data)
        text.docIndex = self._docIndex
        self._docIndex = self._docIndex + 1
        self._curr_node.appendChild(text)
        return

    def processingInstruction(self, target, data):
        pi = ProcessingInstruction(target, data)
        pi.docIndex = self._docIndex
        self._docIndex = self._docIndex + 1
        self._curr_node.appendChild(pi)
        return

    def comment(self, data):
        comment = Comment(data)
        comment.docIndex = self._docIndex
        self._docIndex = self._docIndex + 1
        self._curr_node.appendChild(comment)
        return


# The Result Tree DOM

class RTFNode(Node):
    def __init__(self, nodeType, nodeName, namespaceURI=EMPTY_NAMESPACE, prefix=None, localName=None):
        self.childNodes = []
        self.nodeName = nodeName
        self.nodeValue = None
        self.parentNode = None
        self.firstChild = None
        self.lastChild = None
        self.previousSibling = None
        self.nextSibling = None
        self.attributes = None
        self.ownerDocument = None
        self.namespaceURI = namespaceURI
        self.prefix = prefix
        self.localName = localName
        self.nodeType = nodeType
        return

    def appendChild(self, newChild):
        newChild.parentNode = self
        newChild.ownerDocument = self.ownerDocument or self
        if not self.firstChild:
            self.firstChild = newChild
        else:
            self.lastChild.nextSibling = newChild
            newChild.previousSibling = self.lastChild
        self.lastChild = newChild
        self.childNodes.append(newChild)
        self.stringValue = self.stringValue + newChild.stringValue
        return newChild

class Document(RTFNode):
    def __init__(self):
        RTFNode.__init__(self, Node.DOCUMENT_NODE, u'#document')
        self.stringValue = u''
        return

    def createDocumentFragment(self):
        return DocumentFrag()

    def __repr__(self):
        return '<RTFDocument at %x>' % id(self)


class DocumentFrag(RTFNode):
    def __init__(self):
        RTFNode.__init__(self, Node.DOCUMENT_NODE, u'#document-fragment')
        self.stringValue = u''
        return

    def __repr__(self):
        return '<RTFDocumentFragment at %x>' % id(self)

class Element(RTFNode):
    def __init__(self, nodeName, namespaceURI, prefix, localName):
        RTFNode.__init__(self, Node.ELEMENT_NODE, nodeName, namespaceURI,
                         prefix,localName)
        self.attributes = {}
        self.stringValue = u''

    def setAttributeNodeNS(self, attr):
        attr.ownerDocument = self.ownerDocument
        attr.ownerElement = self
        if attr.localName == 'xmlns':
            self.attributes[(XMLNS_NAMESPACE, attr.prefix)] = attr
        else:
            self.attributes[(attr.namespaceURI, attr.localName)] = attr
        return

    def __repr__(self):
        return "<RTFElement at %x: %s with %d attributes and %d children>" % (
            id(self),
            self.nodeName,
            len(self.attributes),
            len(self.childNodes)
            )

class Attr(RTFNode):
    def __init__(self, nodeName, namespaceURI, prefix, localName, value):
        RTFNode.__init__(self, Node.ATTRIBUTE_NODE, nodeName, namespaceURI,
                      prefix,localName)
        self.ownerElement = None
        self.nodeValue = self.value = value
        self.name = nodeName
        self.stringValue = u''
        self.appendChild(Text(value))

    def __repr__(self):
         return "<RTFAttribute at %x: %s=%s>" % (
             id(self),
             self.name,
             repr(self.value)
             )

class Text(RTFNode):
    def __init__(self, data):
        RTFNode.__init__(self, Node.TEXT_NODE, u'#text')
        self.stringValue = self.nodeValue = self.data = data

    def __repr__(self):
        # Trim to a managable size
        if len(self.data) > 20:
            data = self.data[:20] + '...'
        else:
            data = self.data
        return "<RTFText at %x: %s>" % (id(self), repr(data))

class ProcessingInstruction(RTFNode):
    def __init__(self, target, data):
        RTFNode.__init__(self, Node.PROCESSING_INSTRUCTION_NODE, target)
        self.target = target
        self.stringValue = self.nodeValue = self.data = data

    def __repr__(self):
        if len(self.data) > 20:
            data = self.data[:20] + '...'
        else:
            data = self.data
        return "<RTFProcessingInstruction at %x: %s %s>" % (
            id(self), self.target, data)

class Comment(RTFNode):
    def __init__(self, data):
        RTFNode.__init__(self, Node.COMMENT_NODE, u'#comment')
        self.stringValue = self.nodeValue = self.data = data

    def __repr__(self):
        return "<RTFComment Node at %x: %s>" % (id(self), repr(self.data))

#FIXME: Should no longer need be soo elaborate with GC, right?
def ReleaseRtf(rtf):
    for child in rtf.childNodes:
        ReleaseRtf(child)
    rtf.childNodes = []
    if rtf.nodeType == Node.ELEMENT_NODE:
        for attr in rtf.attributes.values():
            ReleaseRtf(attr)
            attr.ownerElement = None
        rtf.attributes = {}
    rtf.parentNode = None
    rtf.firstChild = None
    rtf.lastChild = None
    rtf.previousSibling = None
    rtf.nextSibling = None
    rtf.ownerDocument = None
    return
