########################################################################
#
# File Name:            Stylesheet.py
#
# Documentation:        http://docs.4suite.org/4XSLT/Stylesheet.py.py.html
#
"""
Implement all the stylesheet internals
WWW: http://4suite.org/4XSLT        e-mail: support@4suite.org

Copyright (c) 1999-2001 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import sys, types
from xml.dom import Node

from Ft.Xml import XPath
from Ft.Xml.XPath import Conversions

from Ft.Xml.Xslt import XSL_NAMESPACE, XsltElement, XsltException, Error
from Ft.Xml.Xslt import CategoryTypes, ContentInfo, AttributeInfo
from Ft.Xml.Xslt import XsltContext, PatternList
from Ft.Xml.Xslt import OutputParameters
from Ft.Xml.Xslt.LiteralElement import LiteralElement

def MatchTree(patterns, context):
    '''Select all nodes from node on down that match the pattern'''
    state = context.copy()

    # Save these before any changes are made to the context
    children = context.node.childNodes
    attributes = context.node.attributes

    matched = patterns.xsltKeyPrep(context, context.node)

    pos = 1
    size = len(children)
    for node in children:
        context.node, context.position, context.size = node, pos, size
        map(lambda x, y: x.extend(y), matched, MatchTree(patterns, context))
        pos += 1
        
    if attributes:
        size = len(attributes)
        pos = 1
        for node in attributes.values():
            context.node, context.position, context.size = node, pos, size
            map(lambda x, y: x.extend(y),
                matched, patterns.xsltKeyPrep(context, node))
            pos += 1

    context.set(state)
    return matched


class StylesheetElement(XsltElement):
    category = None
    content = ContentInfo.Seq(
        ContentInfo.Rep(ContentInfo.QName(XSL_NAMESPACE, 'xsl:import')),
        ContentInfo.TopLevelElements)
    legalAttrs = {
        'id' : AttributeInfo.Id(),
        'extension-element-prefixes' : AttributeInfo.Prefixes(),
        'exclude-result-prefixes' : AttributeInfo.Prefixes(),
        'version' : AttributeInfo.Number(required=1),
        }

    # We don't want ourselves included in these lists since we do the walking
    doesSetup = doesPrime = doesIdle = 0

    def __init__(self, root, namespaceUri, localName, baseUri):
        XsltElement.__init__(self, root, namespaceUri, localName, baseUri)
        self.matchTemplates = {}
        self.namedTemplates = {}
        self.globalVars = {}
        self.initialFunctions = {}
        return

    def setup(self):
        '''
        Called only once, at the first initialization
        '''
        self.outputParams = OutputParameters.OutputParameters()
        self.spaceRules = []
        self.namespaceAliases = {}
        self.decimalFormats = {}

        # By sharing the same list, no special processing is needed later
        space_rules = []
        global_vars = []

        # no need for multiple filters to get these
        top_level_elements = {
            'namespace-alias' : [],
            'decimal-format' : [],
            'strip-space' : space_rules,
            'preserve-space' : space_rules,
            'output' : [],
            'template' : [],
            'key' : [],
            'variable' : global_vars,
            'param' : global_vars,
            }

        # this merges the XSLT children into the above dictionary
        # any elements whose name is not in the dictionary are thrown away
        reduce(lambda ignored, child, toplevel=top_level_elements:
               child.expandedName[0] == XSL_NAMESPACE and
               toplevel.get(child.expandedName[1], []).append(child),
               self.children, 'ignored')

        # process elements which do not depend on other parts of the stylesheet
        self._setupNamespaceAliases(top_level_elements['namespace-alias'])
        self._setupDecimalFormats(top_level_elements['decimal-format'])
        self._setupWhitespaceRules(space_rules)
        self._setupOutput(top_level_elements['output'])
        self._setupTemplates(top_level_elements['template'])
        self._setupKeys(top_level_elements['key'])

        # process those which could require other parts of the stylesheet
        if self.namespaceAliases:
            self._setupLiteralElements(self)

        self._setupTopLevelVarParams(global_vars)
        return

    def _setupNamespaceAliases(self, aliases):
        for alias in aliases:
            stylesheet_ns = self.namespaces[alias._stylesheet_prefix]
            result_ns = self.namespaces[alias._result_prefix]

            namespace_alias = (result_ns, alias._stylesheet_prefix,
                               alias._result_prefix, alias.importIndex)

            if self.namespaceAliases.has_key(stylesheet_ns):
                if self.namespaceAliases[stylesheet_ns][-1] < alias.importIndex:
                    self.namespaceAliases[stylesheet_ns] = namespace_alias
                elif self.namespaceAliases[stylesheet_ns][-1] == alias.importIndex:
                    raise XsltException(Error.DUPLICATE_NAMESPACE_ALIAS,
                                        alias._stylesheet_prefix)
            else:
                self.namespaceAliases[stylesheet_ns] = namespace_alias
        return

    def _setupDecimalFormats(self, decimal_formats):
        for df in decimal_formats:
            name, format = df.getFormatInfo()
            existing = self.decimalFormats.get(name)
            if existing and existing != format:
                # conflicting formats
                if name:
                    # name is (namespace-uri, local-name)
                    if name[0]:
                        name = df.namespaces[name[0]] + ':' + name[1]
                    else:
                        name = name[1]
                else:
                    name = '#default'
                raise XsltException(Error.DUPLICATE_DECIMAL_FORMAT, name)
            self.decimalFormats[name] = format

        if not self.decimalFormats.has_key(None):
            # no default format specified
            self.decimalFormats[None] = ('.', ',', 'Infinity', '-', 'NaN', '%',
                                         unichr(0x2030), '0', '#', ';')
        return

    def _setupWhitespaceRules(self, space_rules):
        # These values only effect the source document(s)
        # Sort in increasing import precedence, so the last one added
        # will have the highest import precedence
        space_rules.sort(lambda a, b: cmp(a.importIndex, b.importIndex))
        merged = {}
        for rule in space_rules:
            strip, elements = rule.getWhitespaceInfo()
            for eName in elements:
                merged[eName] = strip

        # these are sorted by priority where '*' is lowest followed by
        # 'prefix:*', then remaining tests
        star = None
        prefix_star = []
        for (namespace, local), strip in merged.items():
            rule = (namespace, local, strip)
            if local == '*':
                if namespace:
                    prefix_star.append(rule)
                else:
                    star = rule
            else:
                self.spaceRules.append(rule)
        self.spaceRules.extend(prefix_star)
        star and self.spaceRules.append(star)
        return

    def _setupOutput(self, outputs):
        outputs.sort(lambda a, b: cmp(a.importIndex, b.importIndex))
        for output in outputs:
            self.outputParams.parse(output)
        return

    def _setupTemplates(self, templates):
        named_tpls = self.namedTemplates
        match_tpls = self.matchTemplates

        patterns = []
        for template in templates:
            (shortcuts, name) = template.getTemplateInfo()
            if name:
                # for call-templates
                existing = named_tpls.get(name)
                if existing:
                    if existing.importIndex == template.importIndex:
                        raise XsltException(Error.DUPLICATE_NAMED_TEMPLATE, name)
                    elif existing.importIndex > template.importIndex:
                        pass
                    else:
                        # The new template has a higher import precedence
                        # so replace what was already there.
                        named_tpls[name] = template
                else:
                    # The first occurence of a template with this name.
                    named_tpls[name] = template

            patterns.extend(shortcuts)

        def compare(shortcut1, shortcut2):
            # a 'shortcut' is (importidx, priority, mode, patterninfo, quickkey)
            # NOTE: we want higher first, so the comparisions are reversed
            # 1. import precedence
            diff = cmp(shortcut2[0], shortcut1[0])
            if not diff:
                # 2. pattern priority
                diff = cmp(shortcut2[1], shortcut1[1])
                if not diff:
                    # 3. reverse document order
                    # -1 -> arg1 before arg2
                    #  0 -> no change
                    # +1 -> arg2 before arg1
                    diff = +1
            return diff

        patterns.sort(compare)

        # Make the lookup tables
        # A lookup table is first keyed by mode,
        # then keyed by node type,
        # then, if an element, keyed by expanded name
        for j1, j2, mode, pattern_info, (node_type, expanded_name) in patterns:
            mode_table = match_tpls.get(mode)
            if not mode_table:
                mode_table = match_tpls[mode] = {}

            type_table = mode_table.get(node_type)
            if not type_table:
                if node_type == Node.ELEMENT_NODE:
                    # Elements are further keyed by expanded name
                    type_list = [pattern_info]
                    mode_table[node_type] = {expanded_name : type_list}
                else:
                    mode_table[node_type] = [pattern_info]
            elif node_type == Node.ELEMENT_NODE:
                # Elements are further keyed by expanded name
                type_list = type_table.get(expanded_name)
                if not type_list:
                    type_table[expanded_name] = [pattern_info]
                else:
                    type_list.append(pattern_info)
            else:
                # Every other node type gets lumped into a single list
                # for that node type
                type_table.append(pattern_info)
        return

    def _setupKeys(self, keys):
        self._keys = map(lambda key: key.getKeyInfo(), keys)
        return

    def _setupLiteralElements(self, node):
        if isinstance(node, LiteralElement):
            node.fixupAliases(self.namespaceAliases)
        for child in (node.children or []):
            self._setupLiteralElements(child)
        return

    def _setupTopLevelVarParams(self, global_vars):
        self._topVariables = index, ordered = {}, []
        for vp in global_vars:
            existing = index.get(vp._name)
            if vp.importIndex > (existing and existing.importIndex or -1):
                # current variable has a higher import precedence
                index[vp._name] = vp
                ordered.append(vp)
        return

    ############################# Prime Routines #############################

    def prime(self, contextNode, processor, topLevelParams, docUri):
        doc = contextNode.ownerDocument or contextNode
        #self.newSource(doc, processor)

        context = XsltContext.XsltContext(doc, 1, 1,
                                          processorNss=self.namespaces,
                                          stylesheet=self,
                                          processor=processor,
                                          extFunctionMap=processor.extFunctions
                                          )

        baseUri = docUri or getattr(context.node, 'refUri', None)
        context.addDocument(doc, baseUri)

        overridden_params = {}
        for split_name, value in topLevelParams.items():
            if type(split_name) != types.TupleType:
                try:
                    split_name = self.expandQName(split_name)
                except KeyError:
                    continue
            overridden_params[split_name] = value

        for vnode in self._topVariables[1]:
            self._computeGlobalVar(vnode._name, context, [], [],
                                   overridden_params, processor)

        self.globalVars.update(context.varBindings)

        # run prime for all instructions that declared they used it
        for instruction in self.root.primeInstructions:
            instruction.prime(processor, context)

        self.initialFunctions.update(context.functions)
        return


    def _computeGlobalVar(self, vname, context, processed, deferred,
                          overriddenParams, processor):
        vnode = self._topVariables[0][vname]

        if vnode in deferred:
            raise XsltException(Error.CIRCULAR_VAR, vname[0], vname[1])
        if vnode in processed:
            return
        # Is it an <xsl:param>?
        # expandedName is a tuple (namespace-uri, local-name)
        if vnode.expandedName[1][0] == 'p':
            if overriddenParams.has_key(vname):
                context.varBindings[vname] = overriddenParams[vname]
            else:
                finished = 0
                while not finished:
                    try:
                        context = vnode.instantiate(context, processor)[0]
                        finished = 1
                    except XPath.RuntimeException, e:
                        if e.errorCode == XPath.RuntimeException.UNDEFINED_VARIABLE:
                            deferred.append(vnode)
                            self._computeGlobalVar((e.params[0], e.params[1]),
                                                   context, processed, deferred,
                                                   overriddenParams,
                                                   processor)
                            deferred.remove(vnode)
                        else:
                            raise e
                #Set up so that later stylesheets will get overridden by
                #parameter values set in higher-priority stylesheets
                overriddenParams[vname] = context.varBindings[vname]
        else:
            finished = 0
            while not finished:
                try:
                    context = vnode.instantiate(context, processor)[0]
                    finished = 1
                except XPath.RuntimeException, e:
                    if e.errorCode == XPath.RuntimeException.UNDEFINED_VARIABLE:
                        deferred.append(vnode)
                        self._computeGlobalVar((e.params[0], e.params[1]),
                                               context, processed, deferred,
                                               overriddenParams,
                                               processor)
                        deferred.remove(vnode)
                    else:
                        raise e
        processed.append(vnode)
        return

    def updateKeys(self, doc, processor):
        """
        Update the keys for a new document
        """
        processor.keys[doc] = {}
        if not self._keys:
            return

        context = XsltContext.XsltContext(doc, 1, 1,
                                          processorNss=self.namespaces,
                                          processor=processor,
                                          extFunctionMap=self.initialFunctions)
        patterns = map(lambda x: x[1], self._keys)
        patterns = PatternList(patterns, self.namespaces)
        match_lists = MatchTree(patterns, context)
        ctr = 0
        for (name, match_pattern, use_expr) in self._keys:
            match_list = match_lists[ctr]
            if not processor.keys[doc].has_key(name):
                processor.keys[doc][name] = {}
            for node in match_list:
                state = context.copy()
                context.node = node
                key_value_list = use_expr.evaluate(context)
                if not isinstance(key_value_list, types.ListType):
                    key_value_list = [key_value_list]
                for obj in key_value_list:
                    keystr = Conversions.StringValue(obj)
                    if not processor.keys[doc][name].has_key(keystr):
                        processor.keys[doc][name][keystr] = []
                    processor.keys[doc][name][keystr].append(node)
                context.set(state)
            ctr = ctr + 1
        return

    ############################# Exit Routines #############################

    def idle(self, contextNode, processor, baseUri=None):
        # run idle for all instructions that declared they used it
        for instruction in self.root.idleInstructions:
            instruction.idle(processor)
        return

    def reset(self):
        """
        Called whenever the processor is reset, i.e. after each run
        """
        return

    ############################ Runtime Routines ############################

    def getNamedTemplates(self):
        return self.namedTemplates.copy()

    def getGlobalVariables(self):
        return self.globalVars.copy()

    def getInitialFunctions(self):
        return self.initialFunctions.copy()

    def applyTemplates(self, context, processor, params=None,
                       maxImport=sys.maxint):
        # Set the current node for this template application
        context.currentNode = context.node
        context.stylesheet = self

        mode_table = self.matchTemplates.get(context.mode)
        if not mode_table:
            # No patterns defined in this mode
            return 0

        node_type = context.node.nodeType
        if node_type == Node.ELEMENT_NODE:
            table = mode_table.get(node_type)
            if table:
                key = (context.node.namespaceURI, context.node.localName)
                # First use those template which could match this expanded name
                patterns = table.get(key, [])
                # Then add those that are wildcard tests ('*' and 'prefix:*')
                patterns.extend(table.get(None, []))
            else:
                patterns = []
        else:
            patterns = mode_table.get(node_type, [])

        # For any node type add those patterns that don't have a distinct type
        patterns.extend(mode_table.get(None, []))

        for (pattern, axis_type, template) in patterns:
            context.processorNss = template.namespaces
            if template.importIndex < maxImport and \
               pattern.match(context, context.node, axis_type):
                # We successfully match, now use it
                template.instantiate(context, processor, params or {})
                return 1

        # No matching patterns found
        return 0
