########################################################################
#
# File Name:            XmlWriter.py
#
# Documentation:        http://docs.4suite.org/4XSLT/XmlWriter.py.html
#
"""
Implements the XML output writer for XSLT processor output
WWW: http://4suite.org/4XSLT        e-mail: support@4suite.org

Copyright (c) 1999-2001 Fourthought Inc., USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import os, re
from Ft.Xml import EMPTY_NAMESPACE, SplitQName
from Ft.Xml.Lib.String import TranslateCdata, TranslateCdataAttr
from Ft.Xml.Xslt import XSL_NAMESPACE, NullWriter, XsltException, Error
from Ft.Xml import XML_NAMESPACE
from Ft.Xml.Domlette import XmlStrStrip


class ElementData:
    def __init__(self, name, cdataElement, attrs, extraNss=None):
        self.name = name
        self.cdataElement = cdataElement
        self.attrs = attrs
        self.extraNss = extraNss or {}
        return


class XmlWriter(NullWriter.NullWriter):
    GENERATED_PREFIX = u"ns%s"

    def __init__(self, outputParams, stream=None):
        outputParams.setDefault('version', '1.0')
        outputParams.setDefault('encoding', 'UTF-8')
        outputParams.setDefault('indent', 0)
        outputParams.setDefault('mediaType', 'text/xml')
        NullWriter.NullWriter.__init__(self, outputParams, stream)

        self._currElement = None
        self._namespaces = [{'': EMPTY_NAMESPACE, 'xml': XML_NAMESPACE}]
        self._indent = ''
        self._nextNewLine = 0
        self._cdataSectionElement = 0
        self._first_element = 1
        self._cached = []
        self._lastChars = ''
        return

    def _doctype(self, docElem):
        external_id = ''
        if self._outputParams.doctypePublic and self._outputParams.doctypeSystem:
            external_id = u' PUBLIC "' + self._outputParams.doctypePublic + u'" "' + self._outputParams.doctypeSystem + u'"'
        elif self._outputParams.doctypeSystem:
            external_id = u' SYSTEM "' + self._outputParams.doctypeSystem + u'"'
        if external_id:
            self._stream.write(u'<!DOCTYPE %s%s>\n' % (docElem, external_id))
        self._first_element = 0
        return

    def startDocument(self):
        if not self._outputParams.omitXmlDeclaration:
            self._stream.write(u"<?xml version='%s' encoding='%s'" % (
                self._outputParams.version,
                self._outputParams.encoding))
            if self._outputParams.standalone is not None:
                self._stream.write(u" standalone='%s'" %
                                   self._outputParams.standalone)
            self._stream.write(u"?>\n")
        return
        
    def endDocument(self):
        self._completeLastElement(0)
        return

    def text(self, text, escapeOutput=1, asis=0):
        #print "text", repr(text), escapeOutput
        self._completeLastElement(0)
        if escapeOutput:
            if text and text[0] == '>':
                last_chars = self._lastChars
            else:
                last_chars = ''
            text = TranslateCdata(
                text,
                self._outputParams.encoding,
                last_chars,
                markupSafe=self._cdataSectionElement
                )
        else:
            text = TranslateCdata(
                text,
                self._outputParams.encoding,
                '',
                markupSafe=1
                )
        text = unicode(text, self._outputParams.encoding)
        #FIXME: would break under perverse condition where user builds
        #erroneous output one character at a time.
        #True fix requires stateful buffering of the output stream
        self._lastChars = text[-2:]
        self._stream.write(text)
        self._nextNewLine = 0
        return

    def declareNsFromAttr(self, prefix, nsuri):
        if not self._namespaces[-1].has_key(prefix):
            # A new prefix/namespace pair
            self._namespaces[-1][prefix] = nsuri
            self._currElement.extraNss[prefix] = nsuri
        elif self._namespaces[-1][prefix] != nsuri:
            # If the prefix/namespace pair doesn't match what we already
            # have, generate a new prefix
            ctr = 0
            while self._namespaces[-1].has_key(prefix):
                prefix = self.GENERATED_PREFIX % ctr
                if self._namespaces[-1].get(prefix) == nsuri:
                    break
                ctr = ctr + 1
            else:
                self._namespaces[-1][prefix] = nsuri
            self._currElement.extraNss[prefix] = nsuri
        return prefix

    def attribute(self, name, value, namespace=EMPTY_NAMESPACE):
        #print "attribute", name, value, namespace
        if not self._currElement:
            raise XsltException(Error.ATTRIBUTE_ADDED_AFTER_ELEMENT)
        value = TranslateCdataAttr(value, self._outputParams.encoding)
        (prefix, local) = SplitQName(name)
        if namespace != EMPTY_NAMESPACE:
            true_prefix = self.declareNsFromAttr(prefix, namespace)
            if true_prefix != prefix:
                name = u":".join([true_prefix, local])
        else:
            name = local
        self._currElement.attrs[name] = value
        return

    def processingInstruction(self, target, data):
        self._completeLastElement(0)
        target = XmlStrStrip(TranslateCdata(target, self._outputParams.encoding, ''))
        data = XmlStrStrip(TranslateCdata(data, self._outputParams.encoding, ''))
        pi = u'<?%s %s?>' % (target, data)
        if self._outputParams.indent:
            self._stream.write(u"\n%s%s" % (self._indent, pi))
        else:
            self._stream.write(pi)
        self._nextNewLine = 1
        return

    def comment(self, body):
        self._completeLastElement(0)
        comment = u"<!--%s-->" % body
        if self._outputParams.indent:
            self._stream.write(u"\n%s%s" % (self._indent, comment))
        else:
            self._stream.write(comment)
        self._nextNewLine = 1
        return

    def startElement(self, name, namespace=EMPTY_NAMESPACE, extraNss=None):
        #print "startElement", name, namespace, extraNss
        extraNss = extraNss or {}

        self._completeLastElement(0)

        if self._first_element:
            self._doctype(name)

        (prefix, local) = SplitQName(name)
        cdatas_flag = (namespace, local) in self._outputParams.cdataSectionElements
        self._currElement = ElementData(name, cdatas_flag, {}, extraNss)
        self._namespaces.append(self._namespaces[-1].copy())
        self._namespaces[-1][prefix] = namespace
        return

    def endElement(self, name):
        #print "endElement", name
        if self._currElement:
            elementIsEmpty = 1
            self._completeLastElement(1)
        else:
            elementIsEmpty = 0
        if self._outputParams.indent:
            self._indent = self._indent[:-2]
        if self._cdataSectionElement:
            self._stream.write(u']]>')
            self._cdataSectionElement = 0
        if self._outputParams.indent and self._nextNewLine and not elementIsEmpty:
            self._stream.write(u'\n' + self._indent)
        self._stream.write((not elementIsEmpty) and (u'</%s>' % name) or '')
        self._nextNewLine = 1
        del self._namespaces[-1]
        return

    def _completeLastElement(self, elementIsEmpty):
        #print "_completeLastElement", elementIsEmpty, self._currElement
        if self._currElement:
            elem = self._currElement
            if self._outputParams.indent and self._nextNewLine:
                self._stream.write(u'\n' + self._indent)
            self._stream.write(u'<' + elem.name)
            #Handle namespaces
            nss = elem.extraNss
            nss.update(self._namespaces[-1])
            prefixes = nss.keys()
            prefixes.sort()
            for prefix in prefixes:
                ns = nss[prefix]
                prev_ns = self._namespaces[-2].get(prefix)
                if ns != prev_ns:
                    if prefix:
                        self._stream.write(u" xmlns:%s='%s'" % (prefix, ns ))
                    else:
                        if ns == EMPTY_NAMESPACE:
                            ns = u''
                        self._stream.write(u" xmlns='%s'" % ns )
            self._namespaces[-1] = nss

            for (name, (value, delimiter)) in elem.attrs.items():
                self._stream.write(u' %s=%s' % (name, delimiter))
                self._stream.write(value)
                self._stream.write(delimiter)
            if elementIsEmpty:
                self._stream.write(u'/>')
            else:
                self._stream.write(u'>')
                if self._currElement.cdataElement:
                    self._stream.write(u'<![CDATA[')
                    self._cdataSectionElement = 1
            self._nextNewLine = 1

            if self._outputParams.indent:
                self._indent = self._indent + u'  '
            self._currElement = None
        return
